use std::{env::consts::OS, path::{Path, PathBuf}};

use percent_encoding::percent_decode_str;
use regex::Regex;
use ropey::RopeSlice;
use tower_lsp::lsp_types::*;

/// 根据 pos 获取到当前光标所在的字符串，相当于 getWordRangeAtPosition
pub fn get_definition_token(line: &RopeSlice, pos: Position) -> String {
    let mut token = String::new();
    let mut line_iter = line.chars();
    for _ in 0..(line.utf16_cu_to_char(pos.character as usize)) {
        line_iter.next();
    }
    let mut c = line_iter.prev();
    while c.is_some() && (c.unwrap().is_alphanumeric() || c.unwrap() == '_') {
        token.push(c.unwrap());
        c = line_iter.prev();
    }
    token = token.chars().rev().collect();
    line_iter = line.chars();
    for _ in 0..(line.utf16_cu_to_char(pos.character as usize)) {
        line_iter.next();
    }
    let mut c = line_iter.next();
    while c.is_some() && (c.unwrap().is_alphanumeric() || c.unwrap() == '_') {
        token.push(c.unwrap());
        c = line_iter.next();
    }
    token
}

pub fn get_word_range_at_position(line: &RopeSlice, pos: Position, regex: Regex) -> Option<(String, Range)> {
    let mut line_iter = line.chars();

    // 找到 character 这个位置
    for _ in 0..(line.utf16_cu_to_char(pos.character as usize)) {
        line_iter.next();
    }
    let mut c = line_iter.prev();

    let mut token = String::new();
    let mut start_character = pos.character;
    // 往前找
    while c.is_some() && (regex.is_match(&c.unwrap().to_string())) {
        token.push(c.unwrap());
        start_character -= 1;
        c = line_iter.prev();
    }
    token = token.chars().rev().collect();
    line_iter = line.chars();

    for _ in 0..(line.utf16_cu_to_char(pos.character as usize)) {
        line_iter.next();
    }
    let mut c = line_iter.next();
    let mut end_character = pos.character;
    // 往后找
    while c.is_some() && (regex.is_match(&c.unwrap().to_string())) {
        token.push(c.unwrap());
        end_character += 1;
        c = line_iter.next();
    }
    
    if token.len() == 0 {
        return None;
    }

    let range = Range::new(
        Position { line: pos.line, character: start_character },
        Position { line: pos.line, character: end_character }
    );

    Some((token, range))
}

/// 根据 uri 获取 hdl 的 language id
pub fn get_language_id_by_uri(uri: &Url) -> String {
    let path = uri.path();
    let ext_name = std::path::Path::new(path)
        .extension()
        .and_then(std::ffi::OsStr::to_str)
        .unwrap_or("");

    match ext_name {
        "vhd" | "vhdl" | "vho" | "vht" => "vhdl".to_string(),
        "v" | "V" | "vh" | "vl" => "verilog".to_string(),
        "sv" | "svh" => "systemverilog".to_string(),
        _ => "plaintext".to_string()
    }
}

/// 根据基础路径和给出的 path 路径，计算出绝对路径
/// 
/// 如果 path 是绝对路径直接返回
/// 会进行存在性检查，不存在的路径直接返回 None
/// 
/// 例子：
/// 
/// resolve_path("/home/ubuntu/hello.v", "./control.v") -> Some("/home/ubuntu/control.v")
/// 
/// resolve_path("/home/ubuntu/hello.v", "/opt/control.v") -> Some("/opt/control.v")
/// 
pub fn resolve_path(base: &str, path_string: &str) -> Option<PathBuf> {
    let path = Path::new(path_string);
    let abs_path: PathBuf;

    if path.is_absolute() {
        abs_path = PathBuf::from(path);
    } else {
        let base = Path::new(base).parent()?;
        abs_path = base.join(path);
    }
    if abs_path.exists() {
        let path = abs_path.as_path().to_owned();
        return Some(path);
    }
    None
}

/// 把 uri 路径转换成合法的操作系统文件路径
/// 
/// windows 中 
/// 
/// /c%3A/Users/MyCpu.v 转换成 c:/Users/MyCpu.v
pub fn to_escape_path(path: &PathBuf) -> PathBuf {
    let decoded_path = percent_decode_str(path.to_str().unwrap()).decode_utf8_lossy();
    let decoded_path_str = decoded_path.as_ref();
    match OS {
        "windows" => {
            // 去掉开头的斜杠
            let trimmed_path_str = decoded_path_str.trim_start_matches('/');
            PathBuf::from(trimmed_path_str)
        },
        _ => {
            // 其他操作系统（如 Linux）保持原样
            PathBuf::from(decoded_path_str)
        },
    }
}

